# Copyright (c) 2006 L. C. Rees
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#    1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#    2. Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#
#    3. Neither the name of psilib nor the names of its contributors may be used
#       to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''WSGI middleware for caching.'''

import time
import rfc822
from StringIO import StringIO

__all__ = ['WsgiMemoize', 'CacheHeader', 'memoize', 'public', 'private',
    'nocache', 'nostore', 'notransform', 'revalidate', 'proxyrevalidate',
    'maxage', 'smaxage', 'vary', 'modified']

def memoize(cache, **kw):
    '''Decorator for caching.'''
    def decorator(application):
        return WsgiMemoize(application, cache, **kw)
    return decorator

def getinput(environ):
    '''Non-destructively retrieves wsgi.input value.'''
    wsginput = environ['wsgi.input']
    # Non-destructively fetch string value of wsgi.input
    if hasattr(wsginput, 'getvalue'):
        qs = wsginput.getvalue()
    # Otherwise, read and reconstruct wsgi.input
    else:
        # Never read more than content length
        clength = int(environ['CONTENT_LENGTH'])
        qs = wsginput.read(clength)
        environ['wsgi.input'] = StringIO(qs)
    return qs

def expiredate(seconds, value):
    '''Expire date headers for cache control.

    @param seconds Seconds
    @param value Value for Cache-Control header
    '''
    now = time.time()
    return {'Cache-Control':value % seconds, 'Date':rfc822.formatdate(now),
        'Expires':rfc822.formatdate(now + seconds)}

def control(application, value):
    '''Generic setter for 'Cache-Control' headers.

    @param application WSGI application
    @param value 'Cache-Control' value
    '''
    headers = {'Cache-Control':value}
    return CacheHeader(application, headers)

def expire(application, value):
    '''Generic setter for 'Cache-Control' headers + expiration info.

    @param application WSGI application
    @param value 'Cache-Control' value
    '''
    now = rfc822.formatdate()
    headers = {'Cache-Control':value, 'Date':now, 'Expires':now}
    return CacheHeader(application, headers)

def age(value, second):
    '''Generic setter for 'Cache-Control' headers + future expiration info.

    @param value 'Cache-Control' value
    @param seconds # of seconds a resource should be considered invalid in
    '''
    def decorator(application):
        return CacheHeader(application, expiredate(second, value))
    return decorator

def public(application):
    '''Response MAY be cached.'''
    return control(application, 'public')

def private(application):
    '''Response intended for 1 user that MUST NOT be cached.'''
    return expire(application, 'private')

def nocache(application):
    '''Response that a cache can't send without origin server revalidation.'''
    now = rfc822.formatdate()
    headers = {'Cache-Control':'no-cache', 'Pragma':'no-cache', 'Date':now,
        'Expires':now}
    return CacheHeader(application, headers)

def nostore(application):
    '''Response that MUST NOT be cached.'''
    return expire(application, 'no-store')

def notransform(application):
    '''A cache must not modify the Content-Location, Content-MD5, ETag,
    Last-Modified, Expires, Content-Encoding, Content-Range, and Content-Type
    headers.
    '''
    return control(application, 'no-transform')

def revalidate(application):
    '''A cache must revalidate a response with the origin server.'''
    return control(application, 'must-revalidate')

def proxyrevalidate(application):
    '''Shared caches must revalidate a response with the origin server.'''
    return control(application, 'proxy-revalidate')

def maxage(seconds):
    '''Sets the maximum time in seconds a response can be cached.'''
    return age('max-age=%d', seconds)

def smaxage(seconds):
    '''Sets the maximum time in seconds a shared cache can store a response.'''
    return age('s-maxage=%d', seconds)

def expires(seconds):
    '''Sets the time a response expires from the cache (HTTP 1.0).'''
    headers = {'Expires':rfc822.formatdate(time.time() + seconds)}
    def decorator(application):
        return CacheHeader(application, headers)
    return decorator

def vary(headers):
    '''Sets which fields allow a cache to use a response without revalidation.'''
    headers = {'Vary':', '.join(headers)}
    def decorator(application):
        return CacheHeader(application, headers)
    return decorator

def modified(seconds=None):
    '''Sets the time a response was modified.'''
    headers = {'Modified':rfc822.formatdate(seconds)}
    def decorator(application):
        return CacheHeader(application, headers)
    return decorator


class CacheHeader(object):

    '''Controls HTTP Cache Control headers.'''

    def __init__(self, application, headers):
        self.application = application
        self.headers = headers

    def __call__(self, environ, start_response):
        # Restrict cache control to GET and HEAD per HTTP 1.1 RFC
        if environ.get('REQUEST_METHOD') in ('GET', 'HEAD'):
            # Co-routine to add cache control headers
            def hdr_response(status, headers, exc_info=None):
                theaders = self.headers.copy()
                # Aggregate all 'Cache-Control' directives
                if 'Cache-Control' in theaders:
                    for idx, i in enumerate(headers):
                        if i[0] != 'Cache-Control': continue
                        curval = theaders.pop('Cache-Control')
                        newval = ', '.join([curval, i[1]])
                        headers.append(('Cache-Control', newval))
                        del headers[idx]
                        break
                headers.extend((k, v) for k, v in theaders.iteritems())
                return start_response(status, headers, exc_info)
            return self.application(environ, hdr_response)
        return self.application(environ, start_response)


class WsgiMemoize(object):

    '''WSGI middleware for response memoizing.'''

    def __init__(self, app, cache, **kw):
        self.application, self._cache = app, cache
        # Adds method to cache key
        self._methkey = kw.get('key_methods', False)
        # Adds user submitted data to cache key
        self._userkey = kw.get('key_user_info', False)
        # Which HTTP responses by method are cached
        self._allowed = kw.get('allowed_methods', set(['GET', 'HEAD']))

    def __call__(self, environ, start_response):
        # Verify requested response is cacheable
        if environ['REQUEST_METHOD'] not in self._allowed:
            return self.application(environ, start_response)
        # Generate cache key
        key = self._keygen(environ)
        # Query cache for key prescence
        info = self._cache.get(key)
        # Return cached data
        if info is not None:
            start_response(info['status'], info['headers'], info['exc_info'])
            return info['data']
        # Cache start_response info
        def cache_response(status, headers, exc_info=None):
            # Add HTTP cache control headers
            newhdrs = expiredate(self._cache.timeout, 's-maxage=%d')
            headers.extend((k, v) for k, v in newhdrs.iteritems())
            cachedict = {'status':status, 'headers':headers, 'exc_info':exc_info}
            self._cache.set(key, cachedict)
            return start_response(status, headers, exc_info)
        # Wrap data in list to trigger iterator (Roberto De Alemeida)
        data = list(self.application(environ, cache_response))
        # Fetch cached dictionary
        info = self._cache.get(key)
        # Store in dictionary
        info['data'] = data
        # Store in cache
        self._cache.set(key, info)
        # Return data as response to intial request
        return data

    def _keygen(self, environ):
        '''Generates cache keys.'''
        # Base of key is always path of request
        key = [environ['PATH_INFO']]
        # Add method name to key if configured that way
        if self._methkey: key.append(environ['REQUEST_METHOD'])
        # Add user submitted data to string if configured that way
        if self._userkey:
            qs = environ.get('QUERY_STRING', '')
            if qs != '':
                key.append(qs)
            else:
                win = getinput(environ)
                if win != '': key.append(win)
        return ''.join(key)


# Copyright (c) 2005, the Lawrence Journal-World
# Copyright (c) 2006 L. C. Rees
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#    1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#    2. Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#
#    3. Neither the name of Django nor the names of its contributors may be used
#       to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Database cache backend.'''

import time
from random import choice
from datetime import datetime

from sqlalchemy import (
        Table, Column, String, Binary, DateTime,
        Integer, PickleType, bindparam, select, update,
        delete, insert,
    )
try:
    from sqlalchemy import BoundMetaData
except ImportError:
    from sqlalchemy import MetaData as BoundMetaData

from wsgistate import BaseCache
from wsgistate.cache import WsgiMemoize
from wsgistate.session import CookieSession, URLSession, SessionCache

__all__ = ['DbCache', 'memoize', 'session', 'urlsession']

def dbmemo_deploy(global_conf, **kw):
    '''Paste Deploy loader for caching.'''
    def decorator(application):
        _db_memo_cache = DbCache(kw.get('cache'), **kw)
        return WsgiMemoize(application, _db_memo_cache, **kw)
    return decorator

def dbsession_deploy(global_conf, **kw):
    '''Paste Deploy loader for sessions.'''
    def decorator(application):
        _db_base_cache = DbCache(kw.get('cache'), **kw)
        _db_session_cache = SessionCache(_db_base_cache, **kw)
        return CookieSession(application, _db_session_cache, **kw)
    return decorator

def dburlsess_deploy(global_conf, **kw):
    '''Paste Deploy loader for URL encoded sessions.

    @param initstr Database initialization string
    '''
    def decorator(application):
        _db_ubase_cache = DbCache(kw.get('cache'), **kw)
        _db_url_cache = SessionCache(_db_ubase_cache, **kw)
        return URLSession(application, _db_url_cache, **kw)
    return decorator

def memoize(initstr, **kw):
    '''Decorator for caching.

    @param initstr Database initialization string
    '''
    def decorator(application):
        _db_memo_cache = DbCache(initstr, **kw)
        return WsgiMemoize(application, _db_memo_cache, **kw)
    return decorator

def session(initstr, **kw):
    '''Decorator for sessions.

    @param initstr Database initialization string
    '''
    def decorator(application):
        _db_base_cache = DbCache(initstr, **kw)
        _db_session_cache = SessionCache(_db_base_cache, **kw)
        return CookieSession(application, _db_session_cache, **kw)
    return decorator

def urlsession(initstr, **kw):
    '''Decorator for URL encoded sessions.

    @param initstr Database initialization string
    '''
    def decorator(application):
        _db_ubase_cache = DbCache(initstr, **kw)
        _db_url_cache = SessionCache(_db_ubase_cache, **kw)
        return URLSession(application, _db_url_cache, **kw)
    return decorator


class DbCache(BaseCache):

    '''Database cache backend.'''

    def __init__(self, *a, **kw):
        super(DbCache, self).__init__(self, *a, **kw)
        # Get table name
        tablename = kw.get('tablename', 'cache')
        # Bind metadata
        self._metadata = BoundMetaData(a[0])
        # Make cache
        self._cache = Table(
            tablename, self._metadata,
            Column(
                'id', Integer, primary_key=True, nullable=False, unique=True
            ),
            Column('key', String(60), nullable=False),
            Column('value', PickleType, nullable=False),
            Column('expires', DateTime, nullable=False)
        )
        # Create cache if it does not exist
        if not self._cache.exists(): self._cache.create()
        # Maximum number of entries to cull per call if cache is full
        self._maxcull = kw.get('maxcull', 10)
        max_entries = kw.get('max_entries', 300)
        try:
            self._max_entries = int(max_entries)
        except (ValueError, TypeError):
            self._max_entries = 300

    def __len__(self):
        return self._cache.count().execute().fetchone()[0]

    def get(self, key, default=None):
        '''Fetch a given key from the cache.  If the key does not exist, return
        default, which itself defaults to None.

        @param key Keyword of item in cache.
        @param default Default value (default: None)
        '''
        row = select(
            [self._cache.c.value, self._cache.c.expires],
            self._cache.c.key==key
        ).execute().fetchone()
        if row is None:
            return default
        if row.expires < datetime.now().replace(microsecond=0):
            self.delete(key)
            return default
        return row.value

    def set(self, key, value):
        '''Set a value in the cache.

        @param key Keyword of item in cache.
        @param value Value to be inserted in cache.
        '''
        if len(self) > self._max_entries: self._cull()
        timeout, cache = self.timeout, self._cache
        # Get expiration time
        expires = datetime.fromtimestamp(
            time.time() + timeout
        ).replace(microsecond=0)
        #try:
        # Update database if key already present
        if key in self:
            update(
                cache,
                cache.c.key==key,
                dict(value=value, expires=expires),
            ).execute()
        # Insert new key if key not present
        else:
            insert(cache, dict(key=key, value=value, expires=expires)).execute()
        # To be threadsafe, updates/inserts are allowed to fail silently
        #except: pass

    def delete(self, k):
        '''Delete a key from the cache, failing silently.

        @param key Keyword of item in cache.
        '''
        delete(self._cache, self._cache.c.key==k).execute()

    def _cull(self):
        '''Remove items in cache to make more room.'''
        cache, maxcull = self._cache, self._maxcull
        # Remove items that have timed out
        now = datetime.now().replace(microsecond=0)
        delete(cache, cache.c.expires < now).execute()
        # Remove any items over the maximum allowed number in the cache
        if len(self) >= self._max_entries:
            # Upper limit for key query
            ul = maxcull * 2
            # Get list of keys
            keys = [
                i[0] for i in select(
                    [cache.c.key], limit=ul
                ).execute().fetchall()
            ]
            # Get some keys at random
            delkeys = list(choice(keys) for i in xrange(maxcull))
            # Delete keys
            fkeys = tuple({'key':k} for k in delkeys)
            delete(cache, cache.c.key.in_(bindparam('key'))).execute(*fkeys)


# Copyright (c) 2005, the Lawrence Journal-World
# Copyright (c) 2006 L. C. Rees
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#    1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#    2. Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#
#    3. Neither the name of Django nor the names of its contributors may be used
#       to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''File-based cache backend'''

import os, time, urllib
try:
    import cPickle as pickle
except ImportError:
    import pickle

from wsgistate.simple import SimpleCache
from wsgistate.cache import WsgiMemoize
from wsgistate.session import CookieSession, URLSession, SessionCache

__all__ = ['FileCache', 'memoize', 'session', 'urlsession']

def filememo_deploy(global_conf, **kw):
    '''Paste Deploy loader for caching.'''
    def decorator(application):
        _file_memo_cache = FileCache(kw.get('cache'), **kw)
        return WsgiMemoize(application, _file_memo_cache, **kw)
    return decorator

def filesess_deploy(global_conf, **kw):
    '''Paste Deploy loader for sessions.'''
    def decorator(application):
        _file_base_cache = FileCache(kw.get('cache'), **kw)
        _file_session_cache = SessionCache(_file_base_cache, **kw)
        return CookieSession(application, _file_session_cache, **kw)
    return decorator

def fileurlsess_deploy(global_conf, **kw):
    '''Paste Deploy loader for URL encoded sessions.

    @param initstr Database initialization string
    '''
    def decorator(application):
        _file_ubase_cache = FileCache(kw.get('cache'), **kw)
        _file_url_cache = SessionCache(_file_ubase_cache, **kw)
        return URLSession(application, _file_url_cache, **kw)
    return decorator

def memoize(path, **kw):
    '''Decorator for caching.

    @param path Filesystem path
    '''
    def decorator(application):
        _file_memo_cache = FileCache(path, **kw)
        return WsgiMemoize(application, _file_memo_cache, **kw)
    return decorator

def session(path, **kw):
    '''Decorator for sessions.

    @param path Filesystem path
    '''
    def decorator(application):
        _file_base_cache = FileCache(path, **kw)
        _file_session_cache = SessionCache(_file_base_cache, **kw)
        return CookieSession(application, _file_session_cache, **kw)
    return decorator

def urlsession(path, **kw):
    '''Decorator for URL encoded sessions.

    @param path Filesystem path
    '''
    def decorator(application):
        _file_ubase_cache = FileCache(path, **kw)
        _file_url_cache = SessionCache(_file_ubase_cache, **kw)
        return URLSession(application, _file_url_cache, **kw)
    return decorator


class FileCache(SimpleCache):

    '''File-based cache backend'''

    def __init__(self, *a, **kw):
        super(FileCache, self).__init__(*a, **kw)
        # Create directory
        try:
            self._dir = a[0]
        except IndexError:
            raise IOError('file.FileCache requires a valid directory path.')
        if not os.path.exists(self._dir): self._createdir()
        # Remove unneeded methods and attributes
        del self._cache

    def __contains__(self, key):
        '''Tell if a given key is in the cache.'''
        return os.path.exists(self._key_to_file(key))

    def get(self, key, default=None):
        '''Fetch a given key from the cache.  If the key does not exist, return
        default, which itself defaults to None.

        @param key Keyword of item in cache.
        @param default Default value (default: None)
        '''
        try:
            exp, value = pickle.load(open(self._key_to_file(key), 'rb'))
            # Remove item if time has expired.
            if exp < time.time():
                self.delete(key)
                return default
            return value
        except (IOError, OSError, EOFError, pickle.PickleError): pass
        return default

    def set(self, key, value):
        '''Set a value in the cache.

        @param key Keyword of item in cache.
        @param value Value to be inserted in cache.
        '''
        if len(self.keys()) > self._max_entries: self._cull()
        try:
            fname = self._key_to_file(key)
            pickle.dump((time.time() + self.timeout, value), open(fname, 'wb'), 2)
        except (IOError, OSError): pass

    def delete(self, key):
        '''Delete a key from the cache, failing silently.

        @param key Keyword of item in cache.
        '''
        try:
            os.remove(self._key_to_file(key))
        except (IOError, OSError): pass

    def keys(self):
        '''Returns a list of keys in the cache.'''
        return os.listdir(self._dir)

    def _createdir(self):
        '''Creates the cache directory.'''
        try:
            os.makedirs(self._dir)
        except OSError:
            raise EnvironmentError('Cache directory "%s" does not exist and ' \
                'could not be created' % self._dir)

    def _key_to_file(self, key):
        '''Gives the filesystem path for a key.'''
        return os.path.join(self._dir, urllib.quote_plus(key))


# Copyright (c) 2005, the Lawrence Journal-World
# Copyright (c) 2005 Allan Saddi <allan@saddi.com>
# Copyright (c) 2006 L. C. Rees
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#    1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#    2. Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#
#    3. Neither the name of Django nor the names of its contributors may be used
#       to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Thread-safe in-memory cache backend.'''

import copy, time
try:
    import threading
except ImportError:
    import dummy_threading as threading

from wsgistate import synchronized
from wsgistate.simple import SimpleCache
from wsgistate.cache import WsgiMemoize
from wsgistate.session import CookieSession, URLSession, SessionCache

__all__ = ['MemoryCache', 'memoize', 'session', 'urlsession']

def memorymemo_deploy(global_conf, **kw):
    '''Paste Deploy loader for caching.'''
    def decorator(application):
        _memory_memo_cache = MemoryCache(kw.get('cache'), **kw)
        return WsgiMemoize(application, _memory_memo_cache, **kw)
    return decorator

def memorysess_deploy(global_conf, **kw):
    '''Paste Deploy loader for sessions.'''
    def decorator(application):
        _memory_base_cache = MemoryCache(kw.get('cache'), **kw)
        _memory_session_cache = SessionCache(_memory_base_cache, **kw)
        return CookieSession(application, _memory_session_cache, **kw)
    return decorator

def memoryurlsess_deploy(global_conf, **kw):
    '''Paste Deploy loader for URL encoded sessions.

    @param initstr Database initialization string
    '''
    def decorator(application):
        _memory_ubase_cache = MemoryCache(kw.get('cache'), **kw)
        _memory_url_cache = SessionCache(_memory_ubase_cache, **kw)
        return URLSession(application, _memory_url_cache, **kw)
    return decorator

def memoize(**kw):
    '''Decorator for caching.'''
    def decorator(application):
        _mem_memo_cache = MemoryCache(**kw)
        return WsgiMemoize(application, _mem_memo_cache, **kw)
    return decorator

def session(**kw):
    '''Decorator for sessions.'''
    def decorator(application):
        _mem_base_cache = MemoryCache(**kw)
        _mem_session_cache = SessionCache(_mem_base_cache, **kw)
        return CookieSession(application, _mem_session_cache, **kw)
    return decorator

def urlsession(**kw):
    '''Decorator for URL encoded sessions.'''
    def decorator(application):
        _mem_ubase_cache = MemoryCache(**kw)
        _mem_url_cache = SessionCache(_mem_ubase_cache, **kw)
        return URLSession(application, _mem_url_cache, **kw)
    return decorator


class MemoryCache(SimpleCache):

    '''Thread-safe in-memory cache backend.'''

    def __init__(self, *a, **kw):
        super(MemoryCache, self).__init__(*a, **kw)
        self._lock = threading.Condition()

    @synchronized
    def get(self, key, default=None):
        '''Fetch a given key from the cache. If the key does not exist, return
        default, which itself defaults to None.

        @param key Keyword of item in cache.
        @param default Default value (default: None)
        '''
        return copy.deepcopy(super(MemoryCache, self).get(key))

    @synchronized
    def set(self, key, value):
        '''Set a value in the cache.

        @param key Keyword of item in cache.
        @param value Value to be inserted in cache.
        '''
        super(MemoryCache, self).set(key, value)

    @synchronized
    def delete(self, key):
        '''Delete a key from the cache, failing silently.

        @param key Keyword of item in cache.
        '''
        super(MemoryCache, self).delete(key)


# Copyright (c) 2005, the Lawrence Journal-World
# Copyright (c) 2006 L. C. Rees
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#    1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#    2. Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#
#    3. Neither the name of Django nor the names of its contributors may be used
#       to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Memcached cache backend'''

import memcache

from wsgistate import BaseCache
from wsgistate.cache import WsgiMemoize
from wsgistate.session import CookieSession, URLSession, SessionCache

__all__ = ['MemCached', 'memoize', 'session', 'urlsession']

def mcmemo_deploy(global_conf, **kw):
    '''Paste Deploy loader for caching.'''
    def decorator(application):
        _mc_memo_cache = MemCached(kw.get('cache'), **kw)
        return WsgiMemoize(application, _mc_memo_cache, **kw)
    return decorator

def mcsess_deploy(global_conf, **kw):
    '''Paste Deploy loader for sessions.'''
    def decorator(application):
        _mc_base_cache = MemCached(kw.get('cache'), **kw)
        _mc_session_cache = SessionCache(_mc_base_cache, **kw)
        return CookieSession(application, _mc_session_cache, **kw)
    return decorator

def mcurlsess_deploy(global_conf, **kw):
    '''Paste Deploy loader for URL encoded sessions.

    @param initstr Database initialization string
    '''
    def decorator(application):
        _mc_ubase_cache = MemCached(kw.get('cache'), **kw)
        _mc_url_cache = SessionCache(_mc_ubase_cache, **kw)
        return URLSession(application, _mc_url_cache, **kw)
    return decorator

def memoize(path, **kw):
    '''Decorator for caching.

    @param path Client path
    '''
    def decorator(application):
        _mc_memo_cache = MemCached(path, **kw)
        return WsgiMemoize(application, _mc_memo_cache, **kw)
    return decorator

def session(path, **kw):
    '''Decorator for sessions.

    @param path Client path
    '''
    def decorator(application):
        _mc_base_cache = MemCached(path, **kw)
        _mc_session_cache = SessionCache(_mc_base_cache, **kw)
        return CookieSession(application, _mc_session_cache, **kw)
    return decorator

def urlsession(path, **kw):
    '''Decorator for URL encoded sessions.

    @param path Client path
    '''
    def decorator(application):
        _mc_ubase_cache = MemCached(path, **kw)
        _mc_url_cache = SessionCache(_mc_ubase_cache, **kw)
        return URLSession(application, _mc_url_cache, **kw)
    return decorator


class MemCached(BaseCache):

    '''Memcached cache backend'''

    def __init__(self, *a, **kw):
        super(MemCached, self).__init__(*a, **kw)
        self._cache = memcache.Client(a[0].split(';'))

    def get(self, key, default=None):
        '''Fetch a given key from the cache.  If the key does not exist, return
        default, which itself defaults to None.

        @param key Keyword of item in cache.
        @param default Default value (default: None)
        '''
        val = self._cache.get(key)
        if val is None: return default
        return val

    def set(self, key, value):
        '''Set a value in the cache.

        @param key Keyword of item in cache.
        @param value Value to be inserted in cache.
        '''
        self._cache.set(key, value, self.timeout)

    def delete(self, key):
        '''Delete a key from the cache, failing silently.

        @param key Keyword of item in cache.
        '''
        self._cache.delete(key)

    def get_many(self, keys):
        '''Fetch a bunch of keys from the cache.

        Returns a dict mapping each key in keys to its value.  If the given
        key is missing, it will be missing from the response dict.

        @param keys Keywords of items in cache.
        '''
        return self._cache.get_multi(keys)

    def _cull(self):
        '''Stub.'''
        pass


# Copyright (c) 2005, the Lawrence Journal-World
# Copyright (c) 2005 Allan Saddi <allan@saddi.com>
# Copyright (c) 2006 L. C. Rees
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#    1. Redistributions of source code must retain the above copyright notice, 
#       this list of conditions and the following disclaimer.
#    
#    2. Redistributions in binary form must reproduce the above copyright 
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#
#    3. Neither the name of Django nor the names of its contributors may be used
#       to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Base Cache class'''

__all__ = ['BaseCache', 'db', 'file', 'memory', 'memcached', 'session', 'simple', 'cache']

def synchronized(func):
    '''Decorator to lock and unlock a method (Phillip J. Eby).

    @param func Method to decorate
    '''
    def wrapper(self, *__args, **__kw):
        self._lock.acquire()
        try:
            return func(self, *__args, **__kw)
        finally:
            self._lock.release()
    wrapper.__name__ = func.__name__
    wrapper.__dict__ = func.__dict__
    wrapper.__doc__ = func.__doc__
    return wrapper


class BaseCache(object):

    '''Base Cache class.'''    
    
    def __init__(self, *a, **kw):
        super(BaseCache, self).__init__()
        timeout = kw.get('timeout', 300)
        try:
            timeout = int(timeout)
        except (ValueError, TypeError):
            timeout = 300
        self.timeout = timeout

    def __getitem__(self, key):
        '''Fetch a given key from the cache.'''
        return self.get(key)

    def __setitem__(self, key, value):
        '''Set a value in the cache. '''
        self.set(key, value)

    def __delitem__(self, key):
        '''Delete a key from the cache.'''
        self.delete(key) 

    def __contains__(self, key):
        '''Tell if a given key is in the cache.'''
        return self.get(key) is not None

    def get(self, key, default=None):
        '''Fetch a given key from the cache.  If the key does not exist, return
        default, which itself defaults to None.

        @param key Keyword of item in cache.
        @param default Default value (default: None)
        '''
        raise NotImplementedError()

    def set(self, key, value):
        '''Set a value in the cache. 

        @param key Keyword of item in cache.
        @param value Value to be inserted in cache.        
        '''
        raise NotImplementedError()

    def delete(self, key):
        '''Delete a key from the cache, failing silently.

        @param key Keyword of item in cache.
        '''
        raise NotImplementedError()

    def get_many(self, keys):
        '''Fetch a bunch of keys from the cache. Returns a dict mapping each
        key in keys to its value.  If the given key is missing, it will be
        missing from the response dict.

        @param keys Keywords of items in cache.        
        '''
        d = dict()
        for k in keys:
            val = self.get(k)
            if val is not None: d[k] = val
        return d


# Copyright (c) 2005 Allan Saddi <allan@saddi.com>
# Copyright (c) 2005, the Lawrence Journal-World
# Copyright (c) 2006 L. C. Rees
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.

import os
import string
import weakref
import atexit
import cgi
import urllib
import sha
import random
import sys
from Cookie import SimpleCookie
from urllib import quote
try:
    import threading
except ImportError:
    import dummy_threading as threading

from wsgistate import synchronized

__all__ = ['SessionCache', 'SessionManager', 'CookieSession', 'URLSession',
     'session', 'urlsession']

def _shutdown(ref):
    cache = ref()
    if cache is not None: cache.shutdown()

def session(cache, **kw):
    '''Decorator for sessions.'''
    def decorator(application):
        return CookieSession(application, cache, **kw)
    return decorator

def urlsession(cache, **kw):
    '''Decorator for URL encoded sessions.'''
    def decorator(application):
        return URLSession(application, cache, **kw)
    return decorator


class SessionCache(object):

    '''Base class for session cache. You first acquire a session by
    calling create() or checkout(). After using the session, you must call
    checkin(). You must not keep references to sessions outside of a check
    in/check out block. Always obtain a fresh reference.
    '''
    # Would be nice if len(idchars) were some power of 2.
    idchars = '-_'.join([string.digits, string.ascii_letters])
    length = 64

    def __init__(self, cache, **kw):
        self._lock = threading.Condition()
        self.checkedout, self._closed, self.cache = dict(), False, cache
        # Sets if session id is random on every access or not
        self._random = kw.get('random', False)
        self._secret = ''.join(self.idchars[ord(c) % len(self.idchars)]
            for c in os.urandom(self.length))
        # Ensure shutdown is called.
        atexit.register(_shutdown, weakref.ref(self))

    def __del__(self):
        self.shutdown()

    # Public interface.

    @synchronized
    def create(self):
        '''Create a new session with a unique identifier.

        The newly-created session should eventually be released by
        a call to checkin().
        '''
        sid, sess = self.newid(), dict()
        self.cache.set(sid, sess)
        self.checkedout[sid] = sess
        return sid, sess

    @synchronized
    def checkout(self, sid):
        '''Checks out a session for use. Returns the session if it exists,
        otherwise returns None. If this call succeeds, the session
        will be touch()'ed and locked from use by other processes.
        Therefore, it should eventually be released by a call to
        checkin().

        @param sid Session id
        '''
        # If we know it's already checked out, block.
        while sid in self.checkedout: self._lock.wait()
        sess = self.cache.get(sid)
        if sess is not None:
            # Randomize session id if set and remove old session id
            if self._random:
                self.cache.delete(sid)
                sid = self.newid()
            # Put in checkout
            self.checkedout[sid] = sess
            return sid, sess
        return None, None

    @synchronized
    def checkin(self, sid, sess):
        '''Returns the session for use by other threads/processes.

        @param sid Session id
        @param session Session dictionary
        '''
        del self.checkedout[sid]
        self.cache.set(sid, sess)
        self._lock.notify()

    @synchronized
    def shutdown(self):
        '''Clean up outstanding sessions.'''
        if not self._closed:
            # Save or delete any sessions that are still out there.
            for sid, sess in self.checkedout.iteritems():
                self.cache.set(sid, sess)
            self.checkedout.clear()
            self.cache._cull()
            self._closed = True

    # Utilities

    def newid(self):
        'Returns session key that is not being used.'
        sid = None
        for num in xrange(10000):
            sid = sha.new(str(random.randint(0, sys.maxint - 1)) +
              str(random.randint(0, sys.maxint - 1)) + self._secret).hexdigest()
            if sid not in self.cache: break
        return sid


class SessionManager(object):

    '''Session Manager.'''

    def __init__(self, cache, environ, **kw):
        self._cache = cache
        self._fieldname = kw.get('fieldname', '_SID_')
        self._path = kw.get('path', '/')
        self.session = self._sid = self._csid = None
        self.expired = self.current = self.new = self.inurl = False
        self._get(environ)

    def _fromcookie(self, environ):
        '''Attempt to load the associated session using the identifier from
        the cookie.
        '''
        cookie = SimpleCookie(environ.get('HTTP_COOKIE'))
        morsel = cookie.get(self._fieldname, None)
        if morsel is not None:
            self._sid, self.session = self._cache.checkout(morsel.value)
            self._csid = morsel.value
            if self._csid != self._sid: self.new = True

    def _fromquery(self, environ):
        '''Attempt to load the associated session using the identifier from
        the query string.
        '''
        self._qdict = dict(cgi.parse_qsl(environ.get('QUERY_STRING', '')))
        value = self._qdict.get(self._fieldname)
        if value is not None:
            self._sid, self.session = self._cache.checkout(value)
            if self._sid is not None:
                self._csid, self.inurl = value, True
                if self._csid != self._sid: self.current = self.new = True

    def _get(self, environ):
        '''Attempt to associate with an existing Session.'''
        # Try cookie first.
        self._fromcookie(environ)
        # Next, try query string.
        if self.session is None: self._fromquery(environ)
        if self.session is None:
            self._sid, self.session = self._cache.create()
            self.new = True

    def close(self):
        '''Checks session back into session cache.'''
        # Check the session back in and get rid of our reference.
        self._cache.checkin(self._sid, self.session)
        self.session = None

    def setcookie(self, headers):
        '''Sets a cookie header if needed.'''
        cookie, name = SimpleCookie(), self._fieldname
        cookie[name], cookie[name]['path'] = self._sid, self._path
        headers.append(('Set-Cookie', cookie[name].OutputString()))

    def seturl(self, environ):
        '''Encodes session ID in URL, if necessary.'''
        path = ''.join([quote(environ.get('SCRIPT_NAME', '')),
            quote(environ.get('PATH_INFO', ''))])
        # Get query
        if self._qdict:
            self._qdict[self._fieldname] = self._sid
        else:
            self._qdict = {self._fieldname:self._sid}
        return '?'.join([path, urllib.urlencode(self._qdict)])

class _Session(object):

    '''WSGI middleware that adds a session service.'''

    def __init__(self, application, cache, **kw):
        self.application, self.cache, self.kw = application, cache, kw
        # environ key
        self.key = kw.get('key', 'com.saddi.service.session')

    def __call__(self, environ, start_response):
        # New session manager instance each time
        sess = SessionManager(self.cache, environ, **self.kw)
        environ[self.key] = sess
        try:
            # Return intial response if new or session id is random
            if sess.new: return self._initial(environ, start_response)
            return self.application(environ, start_response)
        # Always close session
        finally:
            sess.close()


class CookieSession(_Session):

    '''WSGI middleware that adds a session service in a cookie.'''

    def _initial(self, environ, start_response):
        '''Initial response to a cookie session.'''
        def session_response(status, headers, exc_info=None):
            environ[self.key].setcookie(headers)
            return start_response(status, headers, exc_info)
        return self.application(environ, session_response)


class URLSession(_Session):

    '''WSGI middleware that adds a session service in a URL query string.'''

    def _initial(self, environ, start_response):
        '''Initial response to a query encoded session.'''
        url = environ[self.key].seturl(environ)
        # Redirect to URL with session in query component
        start_response('302 Found', [('location', url)])
        return ['The browser is being redirected to %s' % url]


# Copyright (c) 2006 L. C. Rees
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#    1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#    2. Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#
#    3. Neither the name of psilib nor the names of its contributors may be used
#       to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Unit tests for wsgistate.'''

import unittest
import StringIO
import copy
import os
import time
import urlparse

from wsgistate import *


class TestWsgiState(unittest.TestCase):

    '''Test cases for wsgistate.'''

    def dummy_sr(self, status, headers, exc_info=None):
        return headers

    def my_app(self, environ, start_response):
        session = environ['com.saddi.service.session'].session
        count = session.get('count', 0) + 1
        session['count'] = count
        environ['count'] = count
        headers = start_response('200 OK', [])
        if headers: environ['cookie'] = headers[0][1]
        return environ

    def my_app2(self, environ, start_response):
        session = environ['com.saddi.service.session'].session
        count = session.get('count', 0) + 1
        session['count'] = count
        environ['count'] = count
        headers = start_response('200 OK', [])
        return environ

    def my_app3(self, environ, start_response):
        start_response('200 OK', [])
        return ['passed']

    def test_sc_set_getitem(self):
        '''Tests __setitem__ and __setitem__ on SimpleCache.'''
        testcache = simple.SimpleCache()
        testcache['test'] = 'test'
        self.assertEqual(testcache['test'], 'test')

    def test_sc_set_get(self):
        '''Tests set and get on SimpleCache.'''
        testcache = simple.SimpleCache()
        testcache.set('test', 'test')
        self.assertEqual(testcache.get('test'), 'test')

    def test_sc_delitem(self):
        '''Tests __delitem__ on SimpleCache.'''
        testcache = simple.SimpleCache()
        testcache['test'] = 'test'
        del testcache['test']
        self.assertEqual(testcache.get('test'), None)

    def test_sc_set_delete(self):
        '''Tests delete on SimpleCache.'''
        testcache = simple.SimpleCache()
        testcache.set('test', 'test')
        testcache.delete('test')
        self.assertEqual(testcache.get('test'), None)

    def test_set_getmany(self):
        '''Tests delete on SimpleCache.'''
        testcache = simple.SimpleCache()
        testcache.set('test', 'test')
        testcache.set('test2', 'test2')
        self.assertEqual(sorted(testcache.get_many(('test', 'test2')).values()), ['test', 'test2'])

    def test_sc_in_true(self):
        '''Tests in (true) on SimpleCache.'''
        testcache = simple.SimpleCache()
        testcache.set('test', 'test')
        self.assertEqual('test' in testcache, True)

    def test_sc_in_false(self):
        '''Tests in (false) on SimpleCache.'''
        testcache = simple.SimpleCache()
        testcache.set('test2', 'test')
        self.assertEqual('test' in testcache, False)

    def test_sc_timeout(self):
        '''Tests timeout in SimpleCache.'''
        testcache = simple.SimpleCache(timeout=1)
        testcache.set('test', 'test')
        time.sleep(1)
        self.assertEqual(testcache.get('test'), None)

    def test_mc_set_getitem(self):
        '''Tests __setitem__ and __setitem__ on MemoryCache.'''
        testcache = memory.MemoryCache()
        testcache['test'] = 'test'
        self.assertEqual(testcache['test'], 'test')

    def test_mc_set_get(self):
        '''Tests set and get on MemoryCache.'''
        testcache = memory.MemoryCache()
        testcache.set('test', 'test')
        self.assertEqual(testcache.get('test'), 'test')

    def test_mc_delitem(self):
        '''Tests __delitem__ on MemoryCache.'''
        testcache = memory.MemoryCache()
        testcache['test'] = 'test'
        del testcache['test']
        self.assertEqual(testcache.get('test'), None)

    def test_mc_set_delete(self):
        '''Tests delete on MemoryCache.'''
        testcache = memory.MemoryCache()
        testcache.set('test', 'test')
        testcache.delete('test')
        self.assertEqual(testcache.get('test'), None)

    def test_mc_set_getmany(self):
        '''Tests delete on MemoryCache.'''
        testcache = memory.MemoryCache()
        testcache.set('test', 'test')
        testcache.set('test2', 'test2')
        self.assertEqual(sorted(testcache.get_many(('test', 'test2')).values()), ['test', 'test2'])

    def test_mc_in_true(self):
        '''Tests in (true) on MemoryCache.'''
        testcache = memory.MemoryCache()
        testcache.set('test', 'test')
        self.assertEqual('test' in testcache, True)

    def test_mc_in_false(self):
        '''Tests in (false) on MemoryCache.'''
        testcache = memory.MemoryCache()
        testcache.set('test2', 'test')
        self.assertEqual('test' in testcache, False)

    def test_mc_timeout(self):
        '''Tests timeout in MemoryCache.'''
        testcache = memory.MemoryCache(timeout=1)
        testcache.set('test', 'test')
        time.sleep(1)
        self.assertEqual(testcache.get('test'), None)

    def test_fc_set_getitem(self):
        '''Tests __setitem__ and __setitem__ on FileCache.'''
        testcache = file.FileCache('test_wsgistate')
        testcache['test'] = 'test'
        self.assertEqual(testcache['test'], 'test')

    def test_fc_set_get(self):
        '''Tests set and get on FileCache.'''
        testcache = file.FileCache('test_wsgistate')
        testcache.set('test', 'test')
        self.assertEqual(testcache.get('test'), 'test')

    def test_fc_delitem(self):
        '''Tests __delitem__ on FileCache.'''
        testcache = file.FileCache('test_wsgistate')
        testcache['test'] = 'test'
        del testcache['test']
        self.assertEqual(testcache.get('test'), None)

    def test_fc_set_delete(self):
        '''Tests delete on FileCache.'''
        testcache = file.FileCache('test_wsgistate')
        testcache.set('test', 'test')
        testcache.delete('test')
        self.assertEqual(testcache.get('test'), None)

    def test_fc_set_getmany(self):
        '''Tests delete on FileCache.'''
        testcache = file.FileCache('test_wsgistate')
        testcache.set('test', 'test')
        testcache.set('test2', 'test2')
        self.assertEqual(sorted(testcache.get_many(('test', 'test2')).values()), ['test', 'test2'])

    def test_fc_in_true(self):
        '''Tests in (true) on FileCache.'''
        testcache = file.FileCache('test_wsgistate')
        testcache.set('test', 'test')
        self.assertEqual('test' in testcache, True)

    def test_fc_in_false(self):
        '''Tests in (false) on FileCache.'''
        testcache = file.FileCache('test_wsgistate')
        testcache.set('test2', 'test')
        self.assertEqual('test' in testcache, False)

    def test_fc_timeout(self):
        '''Tests timeout in FileCache.'''
        testcache = file.FileCache('test_wsgistate', timeout=1)
        testcache.set('test', 'test')
        time.sleep(1)
        self.assertEqual(testcache.get('test'), None)

    def test_db_set_getitem(self):
        '''Tests __setitem__ and __setitem__ on DbCache.'''
        testcache = db.DbCache('sqlite://')
        testcache['test'] = 'test'
        self.assertEqual(testcache['test'], 'test')

    def test_db_set_get(self):
        '''Tests set and get on DbCache.'''
        testcache = db.DbCache('sqlite://')
        testcache.set('test', 'test')
        self.assertEqual(testcache.get('test'), 'test')

    def test_db_delitem(self):
        '''Tests __delitem__ on DbCache.'''
        testcache = db.DbCache('sqlite://')
        testcache['test'] = 'test'
        del testcache['test']
        self.assertEqual(testcache.get('test'), None)

    def test_db_set_delete(self):
        '''Tests delete on DbCache.'''
        testcache = db.DbCache('sqlite://')
        testcache.set('test', 'test')
        testcache.delete('test')
        self.assertEqual(testcache.get('test'), None)

    def test_db_set_getmany(self):
        '''Tests delete on DbCache.'''
        testcache = db.DbCache('sqlite://')
        testcache.set('test', 'test')
        testcache.set('test2', 'test2')
        self.assertEqual(
            sorted(testcache.get_many(('test', 'test2'))),
                ['test', 'test2']
        )

    def test_db_in_true(self):
        '''Tests in (true) on DbCache.'''
        testcache = db.DbCache('sqlite://')
        testcache.set('test', 'test')
        self.assertEqual('test' in testcache, True)

    def test_db_in_false(self):
        '''Tests in (false) on DbCache.'''
        testcache = db.DbCache('sqlite://')
        testcache.set('test2', 'flasmburbert')
        self.assertEqual('flasmburbert' in testcache, False)

    def test_db_timeout(self):
        '''Tests timeout in DbCache.'''
        testcache = db.DbCache('sqlite://', timeout=1)
        testcache.set('test', 'test')
        time.sleep(2)
        self.assertEqual(testcache.get('test'), None)

    def test_mcd_set_getitem(self):
        '''Tests __setitem__ and __setitem__ on MemCache.'''
        testcache = memcached.MemCached('localhost')
        testcache['test'] = 'test'
        self.assertEqual(testcache['test'], 'test')

    def test_mcd_set_get(self):
        '''Tests set and get on MemCache.'''
        testcache = memcached.MemCached('localhost')
        testcache.set('test', 'test')
        self.assertEqual(testcache.get('test'), 'test')

    def test_mcd_delitem(self):
        '''Tests __delitem__ on MemCache.'''
        testcache = memcached.MemCached('localhost')
        testcache['test'] = 'test'
        del testcache['test']
        self.assertEqual(testcache.get('test'), None)

    def test_mcd_set_delete(self):
        '''Tests delete on MemCache.'''
        testcache = memcached.MemCached('localhost')
        testcache.set('test', 'test')
        testcache.delete('test')
        self.assertEqual(testcache.get('test'), None)

    def test_mcd_set_getmany(self):
        '''Tests delete on MemCache.'''
        testcache = memcached.MemCached('localhost')
        testcache.set('test', 'test')
        testcache.set('test2', 'test2')
        self.assertEqual(sorted(testcache.get_many(('test', 'test2')).values()), ['test', 'test2'])

    def test_mcd_in_true(self):
        '''Tests in (true) on MemCache.'''
        testcache = memcached.MemCached('localhost')
        testcache.set('test', 'test')
        self.assertEqual('test' in testcache, True)

    def test_mcd_in_false(self):
        '''Tests in (false) on MemCache.'''
        testcache = memcached.MemCached('localhost')
        testcache.set('test2', 'test')
        self.assertEqual('test' in testcache, False)

    def test_mcb_timeout(self):
        '''Tests timeout in DbCache.'''
        testcache = memcached.MemCached('localhost', timeout=1)
        testcache.set('test', 'test')
        time.sleep(1)
        self.assertEqual(testcache.get('test'), None)

    def test_cookiesession_sc(self):
        '''Tests session cookies with SimpleCache.'''
        testc = simple.SimpleCache()
        testcache = session.SessionCache(testc)
        csession = session.CookieSession(self.my_app, testcache)
        cookie = csession({}, self.dummy_sr)['cookie']
        csession({'HTTP_COOKIE':cookie}, self.dummy_sr)
        csession({'HTTP_COOKIE':cookie}, self.dummy_sr)
        result = csession({'HTTP_COOKIE':cookie}, self.dummy_sr)
        self.assertEqual(result['count'], 4)

    def test_cookiesession_mc(self):
        '''Tests session cookies with MemoryCache.'''
        testc = memory.MemoryCache()
        testcache = session.SessionCache(testc)
        csession = session.CookieSession(self.my_app, testcache)
        cookie = csession({}, self.dummy_sr)['cookie']
        csession({'HTTP_COOKIE':cookie}, self.dummy_sr)
        csession({'HTTP_COOKIE':cookie}, self.dummy_sr)
        result = csession({'HTTP_COOKIE':cookie}, self.dummy_sr)
        self.assertEqual(result['count'], 4)

    def test_cookiesession_fc(self):
        '''Tests session cookies with FileCache.'''
        testc = file.FileCache('test_wsgistate')
        testcache = session.SessionCache(testc)
        csession = session.CookieSession(self.my_app, testcache)
        cookie = csession({}, self.dummy_sr)['cookie']
        csession({'HTTP_COOKIE':cookie}, self.dummy_sr)
        csession({'HTTP_COOKIE':cookie}, self.dummy_sr)
        result = csession({'HTTP_COOKIE':cookie}, self.dummy_sr)
        self.assertEqual(result['count'], 4)

    def test_cookiesession_dc(self):
        '''Tests session cookies with DbCache.'''
        testc = db.DbCache('sqlite://')
        testcache = session.SessionCache(testc)
        csession = session.CookieSession(self.my_app, testcache)
        cookie = csession({}, self.dummy_sr)['cookie']
        csession({'HTTP_COOKIE':cookie}, self.dummy_sr)
        csession({'HTTP_COOKIE':cookie}, self.dummy_sr)
        result = csession({'HTTP_COOKIE':cookie}, self.dummy_sr)
        self.assertEqual(result['count'], 4)

    def test_cookiesession_mdc(self):
        '''Tests session cookies with MemCached.'''
        testc = memcached.MemCached('localhost')
        testcache = session.SessionCache(testc)
        csession = session.CookieSession(self.my_app, testcache)
        cookie = csession({}, self.dummy_sr)['cookie']
        csession({'HTTP_COOKIE':cookie}, self.dummy_sr)
        csession({'HTTP_COOKIE':cookie}, self.dummy_sr)
        result = csession({'HTTP_COOKIE':cookie}, self.dummy_sr)
        self.assertEqual(result['count'], 4)

    def test_dec_cookiesession_sc(self):
        '''Tests session cookies with SimpleCache decorator.'''
        @simple.session()
        def tapp(environ, start_response):
            session = environ['com.saddi.service.session'].session
            count = session.get('count', 0) + 1
            session['count'] = count
            environ['count'] = count
            headers = start_response('200 OK', [])
            if headers: environ['cookie'] = headers[0][1]
            return environ
        cookie = tapp({}, self.dummy_sr)['cookie']
        tapp({'HTTP_COOKIE':cookie}, self.dummy_sr)
        tapp({'HTTP_COOKIE':cookie}, self.dummy_sr)
        result = tapp({'HTTP_COOKIE':cookie}, self.dummy_sr)
        self.assertEqual(result['count'], 4)

    def test_dec_cookiesession_memc(self):
        '''Tests session cookies with MemoryCache decorator.'''
        @memory.session()
        def tapp(environ, start_response):
            session = environ['com.saddi.service.session'].session
            count = session.get('count', 0) + 1
            session['count'] = count
            environ['count'] = count
            headers = start_response('200 OK', [])
            if headers: environ['cookie'] = headers[0][1]
            return environ
        cookie = tapp({}, self.dummy_sr)['cookie']
        tapp({'HTTP_COOKIE':cookie}, self.dummy_sr)
        tapp({'HTTP_COOKIE':cookie}, self.dummy_sr)
        result = tapp({'HTTP_COOKIE':cookie}, self.dummy_sr)
        self.assertEqual(result['count'], 4)

    def test_dec_cookiesession_fc(self):
        '''Tests session cookies with FileCache decorator.'''
        @file.session('test_wsgistate')
        def tapp(environ, start_response):
            session = environ['com.saddi.service.session'].session
            count = session.get('count', 0) + 1
            session['count'] = count
            environ['count'] = count
            headers = start_response('200 OK', [])
            if headers: environ['cookie'] = headers[0][1]
            return environ
        cookie = tapp({}, self.dummy_sr)['cookie']
        tapp({'HTTP_COOKIE':cookie}, self.dummy_sr)
        tapp({'HTTP_COOKIE':cookie}, self.dummy_sr)
        result = tapp({'HTTP_COOKIE':cookie}, self.dummy_sr)
        self.assertEqual(result['count'], 4)

    def test_dec_cookiesession_db(self):
        '''Tests session cookies with DbCache decorator.'''
        @db.session('sqlite://')
        def tapp(environ, start_response):
            session = environ['com.saddi.service.session'].session
            count = session.get('count', 0) + 1
            session['count'] = count
            environ['count'] = count
            headers = start_response('200 OK', [])
            if headers: environ['cookie'] = headers[0][1]
            return environ
        cookie = tapp({}, self.dummy_sr)['cookie']
        tapp({'HTTP_COOKIE':cookie}, self.dummy_sr)
        tapp({'HTTP_COOKIE':cookie}, self.dummy_sr)
        result = tapp({'HTTP_COOKIE':cookie}, self.dummy_sr)
        self.assertEqual(result['count'], 4)

    def test_dec_cookiesession_mc(self):
        '''Tests session cookies with memcached decorator.'''
        @memcached.session('localhost')
        def tapp(environ, start_response):
            session = environ['com.saddi.service.session'].session
            count = session.get('count', 0) + 1
            session['count'] = count
            environ['count'] = count
            headers = start_response('200 OK', [])
            if headers: environ['cookie'] = headers[0][1]
            return environ
        cookie = tapp({}, self.dummy_sr)['cookie']
        tapp({'HTTP_COOKIE':cookie}, self.dummy_sr)
        tapp({'HTTP_COOKIE':cookie}, self.dummy_sr)
        result = tapp({'HTTP_COOKIE':cookie}, self.dummy_sr)
        self.assertEqual(result['count'], 4)

    def test_random_cookiesession_sc(self):
        '''Tests random session cookies with SimpleCache.'''
        testc = simple.SimpleCache()
        testcache = session.SessionCache(testc, random=True)
        csession = session.CookieSession(self.my_app, testcache)
        result = csession({}, self.dummy_sr)
        result = csession({'HTTP_COOKIE':result['cookie']}, self.dummy_sr)
        result = csession({'HTTP_COOKIE':result['cookie']}, self.dummy_sr)
        result = csession({'HTTP_COOKIE':result['cookie']}, self.dummy_sr)
        self.assertEqual(result['count'], 4)

    def test_random_cookiesession_mc(self):
        '''Tests random session cookies with MemoryCache.'''
        testc = memory.MemoryCache()
        testcache = session.SessionCache(testc, random=True)
        csession = session.CookieSession(self.my_app, testcache)
        result = csession({}, self.dummy_sr)
        result = csession({'HTTP_COOKIE':result['cookie']}, self.dummy_sr)
        result = csession({'HTTP_COOKIE':result['cookie']}, self.dummy_sr)
        result = csession({'HTTP_COOKIE':result['cookie']}, self.dummy_sr)
        self.assertEqual(result['count'], 4)

    def test_random_cookiesession_fc(self):
        '''Tests random session cookies with FileCache.'''
        testc = file.FileCache('test_wsgistate')
        testcache = session.SessionCache(testc, random=True)
        csession = session.CookieSession(self.my_app, testcache)
        result = csession({}, self.dummy_sr)
        result = csession({'HTTP_COOKIE':result['cookie']}, self.dummy_sr)
        result = csession({'HTTP_COOKIE':result['cookie']}, self.dummy_sr)
        result = csession({'HTTP_COOKIE':result['cookie']}, self.dummy_sr)
        self.assertEqual(result['count'], 4)

    def test_random_cookiesession_dc(self):
        '''Tests random session cookies with DbCache.'''
        testc = db.DbCache('sqlite://')
        testcache = session.SessionCache(testc, random=True)
        csession = session.CookieSession(self.my_app, testcache)
        result = csession({}, self.dummy_sr)
        result = csession({'HTTP_COOKIE':result['cookie']}, self.dummy_sr)
        result = csession({'HTTP_COOKIE':result['cookie']}, self.dummy_sr)
        result = csession({'HTTP_COOKIE':result['cookie']}, self.dummy_sr)
        self.assertEqual(result['count'], 4)

    def test_random_cookiesession_mdc(self):
        '''Tests random session cookies with MemCached.'''
        testc = memcached.MemCached('localhost')
        testcache = session.SessionCache(testc, random=True)
        csession = session.CookieSession(self.my_app, testcache)
        result = csession({}, self.dummy_sr)
        result = csession({'HTTP_COOKIE':result['cookie']}, self.dummy_sr)
        result = csession({'HTTP_COOKIE':result['cookie']}, self.dummy_sr)
        result = csession({'HTTP_COOKIE':result['cookie']}, self.dummy_sr)
        self.assertEqual(result['count'], 4)

    def test_dec_random_cookiesession_sc(self):
        '''Tests random session cookies with SimpleCache decorator.'''
        @simple.session(random=True)
        def tapp(environ, start_response):
            session = environ['com.saddi.service.session'].session
            count = session.get('count', 0) + 1
            session['count'] = count
            environ['count'] = count
            headers = start_response('200 OK', [])
            if headers: environ['cookie'] = headers[0][1]
            return environ
        result = tapp({}, self.dummy_sr)
        result = tapp({'HTTP_COOKIE':result['cookie']}, self.dummy_sr)
        result = tapp({'HTTP_COOKIE':result['cookie']}, self.dummy_sr)
        result = tapp({'HTTP_COOKIE':result['cookie']}, self.dummy_sr)
        self.assertEqual(result['count'], 4)

    def test_dec_random_cookiesession_mc(self):
        '''Tests random session cookies with MemoryCache decorator.'''
        @memory.session(random=True)
        def tapp(environ, start_response):
            session = environ['com.saddi.service.session'].session
            count = session.get('count', 0) + 1
            session['count'] = count
            environ['count'] = count
            headers = start_response('200 OK', [])
            if headers: environ['cookie'] = headers[0][1]
            return environ
        result = tapp({}, self.dummy_sr)
        result = tapp({'HTTP_COOKIE':result['cookie']}, self.dummy_sr)
        result = tapp({'HTTP_COOKIE':result['cookie']}, self.dummy_sr)
        result = tapp({'HTTP_COOKIE':result['cookie']}, self.dummy_sr)
        self.assertEqual(result['count'], 4)

    def test_dec_random_cookiesession_fc(self):
        '''Tests random session cookies with FileCache decorator.'''
        @file.session('test_wsgistate', random=True)
        def tapp(environ, start_response):
            session = environ['com.saddi.service.session'].session
            count = session.get('count', 0) + 1
            session['count'] = count
            environ['count'] = count
            headers = start_response('200 OK', [])
            if headers: environ['cookie'] = headers[0][1]
            return environ
        result = tapp({}, self.dummy_sr)
        result = tapp({'HTTP_COOKIE':result['cookie']}, self.dummy_sr)
        result = tapp({'HTTP_COOKIE':result['cookie']}, self.dummy_sr)
        result = tapp({'HTTP_COOKIE':result['cookie']}, self.dummy_sr)
        self.assertEqual(result['count'], 4)

    def test_dec_random_cookiesession_db(self):
        '''Tests random session cookies with DbCache decorator.'''
        @db.session('sqlite://', random=True)
        def tapp(environ, start_response):
            session = environ['com.saddi.service.session'].session
            count = session.get('count', 0) + 1
            session['count'] = count
            environ['count'] = count
            headers = start_response('200 OK', [])
            if headers: environ['cookie'] = headers[0][1]
            return environ
        result = tapp({}, self.dummy_sr)
        result = tapp({'HTTP_COOKIE':result['cookie']}, self.dummy_sr)
        result = tapp({'HTTP_COOKIE':result['cookie']}, self.dummy_sr)
        result = tapp({'HTTP_COOKIE':result['cookie']}, self.dummy_sr)
        self.assertEqual(result['count'], 4)

    def test_dec_random_cookiesession_mcd(self):
        '''Tests random session cookies with memcached decorator.'''
        @memcached.session('localhost', random=True)
        def tapp(environ, start_response):
            session = environ['com.saddi.service.session'].session
            count = session.get('count', 0) + 1
            session['count'] = count
            environ['count'] = count
            headers = start_response('200 OK', [])
            if headers: environ['cookie'] = headers[0][1]
            return environ
        result = tapp({}, self.dummy_sr)
        result = tapp({'HTTP_COOKIE':result['cookie']}, self.dummy_sr)
        result = tapp({'HTTP_COOKIE':result['cookie']}, self.dummy_sr)
        result = tapp({'HTTP_COOKIE':result['cookie']}, self.dummy_sr)
        self.assertEqual(result['count'], 4)

    def test_urlsession_sc(self):
        '''Tests URL encoded sessions with SimpleCache.'''
        testc = simple.SimpleCache()
        testcache = session.SessionCache(testc)
        csession = session.URLSession(self.my_app2, testcache)
        url = csession({}, self.dummy_sr)[0].split()[-1]
        query = urlparse.urlsplit(url)[3]
        csession({'QUERY_STRING':query}, self.dummy_sr)
        csession({'QUERY_STRING':query}, self.dummy_sr)
        csession({'QUERY_STRING':query}, self.dummy_sr)
        result = csession({'QUERY_STRING':query}, self.dummy_sr)
        self.assertEqual(result['count'], 4)

    def test_urlsession_mc(self):
        '''Tests URL encoded sessions with MemoryCache.'''
        testc = memory.MemoryCache()
        testcache = session.SessionCache(testc)
        csession = session.URLSession(self.my_app2, testcache)
        url = csession({}, self.dummy_sr)[0].split()[-1]
        query = urlparse.urlsplit(url)[3]
        csession({'QUERY_STRING':query}, self.dummy_sr)
        csession({'QUERY_STRING':query}, self.dummy_sr)
        csession({'QUERY_STRING':query}, self.dummy_sr)
        result = csession({'QUERY_STRING':query}, self.dummy_sr)
        self.assertEqual(result['count'], 4)

    def test_urlsession_fc(self):
        '''Tests URL encoded sessions with FileCache.'''
        testc = file.FileCache('test_wsgistate')
        testcache = session.SessionCache(testc)
        csession = session.URLSession(self.my_app2, testcache)
        url = csession({}, self.dummy_sr)[0].split()[-1]
        query = urlparse.urlsplit(url)[3]
        csession({'QUERY_STRING':query}, self.dummy_sr)
        csession({'QUERY_STRING':query}, self.dummy_sr)
        csession({'QUERY_STRING':query}, self.dummy_sr)
        result = csession({'QUERY_STRING':query}, self.dummy_sr)
        self.assertEqual(result['count'], 4)

    def test_urlsession_dc(self):
        '''Tests URL encoded sessions with DbCache.'''
        testc = db.DbCache('sqlite://')
        testcache = session.SessionCache(testc)
        csession = session.URLSession(self.my_app2, testcache)
        url = csession({}, self.dummy_sr)[0].split()[-1]
        query = urlparse.urlsplit(url)[3]
        csession({'QUERY_STRING':query}, self.dummy_sr)
        csession({'QUERY_STRING':query}, self.dummy_sr)
        csession({'QUERY_STRING':query}, self.dummy_sr)
        result = csession({'QUERY_STRING':query}, self.dummy_sr)
        self.assertEqual(result['count'], 4)

    def test_urlsession_mdc(self):
        '''Tests URL encoded sessions with MemCached.'''
        testc = memcached.MemCached('localhost')
        testcache = session.SessionCache(testc)
        csession = session.URLSession(self.my_app2, testcache)
        url = csession({}, self.dummy_sr)[0].split()[-1]
        query = urlparse.urlsplit(url)[3]
        csession({'QUERY_STRING':query}, self.dummy_sr)
        csession({'QUERY_STRING':query}, self.dummy_sr)
        csession({'QUERY_STRING':query}, self.dummy_sr)
        result = csession({'QUERY_STRING':query}, self.dummy_sr)
        self.assertEqual(result['count'], 4)

    def test_dec_urlsession_sc(self):
        '''Tests URL encoded sessions with SimpleCache.'''
        @simple.urlsession()
        def csession(environ, start_response):
            session = environ['com.saddi.service.session'].session
            count = session.get('count', 0) + 1
            session['count'] = count
            environ['count'] = count
            headers = start_response('200 OK', [])
            if headers: environ['cookie'] = headers[0][1]
            return environ
        url = csession({}, self.dummy_sr)[0].split()[-1]
        query = urlparse.urlsplit(url)[3]
        csession({'QUERY_STRING':query}, self.dummy_sr)
        csession({'QUERY_STRING':query}, self.dummy_sr)
        csession({'QUERY_STRING':query}, self.dummy_sr)
        result = csession({'QUERY_STRING':query}, self.dummy_sr)
        self.assertEqual(result['count'], 4)

    def test_dec_urlsession_mc(self):
        '''Tests URL encoded sessions with MemoryCache decorator.'''
        @memory.urlsession()
        def csession(environ, start_response):
            session = environ['com.saddi.service.session'].session
            count = session.get('count', 0) + 1
            session['count'] = count
            environ['count'] = count
            headers = start_response('200 OK', [])
            if headers: environ['cookie'] = headers[0][1]
            return environ
        url = csession({}, self.dummy_sr)[0].split()[-1]
        query = urlparse.urlsplit(url)[3]
        csession({'QUERY_STRING':query}, self.dummy_sr)
        csession({'QUERY_STRING':query}, self.dummy_sr)
        csession({'QUERY_STRING':query}, self.dummy_sr)
        result = csession({'QUERY_STRING':query}, self.dummy_sr)
        self.assertEqual(result['count'], 4)

    def test_dec_urlsession_fc(self):
        '''Tests URL encoded sessions with FileCache decorator.'''
        @file.urlsession('test_wsgistate')
        def csession(environ, start_response):
            session = environ['com.saddi.service.session'].session
            count = session.get('count', 0) + 1
            session['count'] = count
            environ['count'] = count
            headers = start_response('200 OK', [])
            if headers: environ['cookie'] = headers[0][1]
            return environ
        url = csession({}, self.dummy_sr)[0].split()[-1]
        query = urlparse.urlsplit(url)[3]
        csession({'QUERY_STRING':query}, self.dummy_sr)
        csession({'QUERY_STRING':query}, self.dummy_sr)
        csession({'QUERY_STRING':query}, self.dummy_sr)
        result = csession({'QUERY_STRING':query}, self.dummy_sr)
        self.assertEqual(result['count'], 4)

    def test_dec_urlsession_dc(self):
        '''Tests URL encoded sessions with DbCache decorator.'''
        @db.urlsession('sqlite://')
        def csession(environ, start_response):
            session = environ['com.saddi.service.session'].session
            count = session.get('count', 0) + 1
            session['count'] = count
            environ['count'] = count
            headers = start_response('200 OK', [])
            if headers: environ['cookie'] = headers[0][1]
            return environ
        url = csession({}, self.dummy_sr)[0].split()[-1]
        query = urlparse.urlsplit(url)[3]
        csession({'QUERY_STRING':query}, self.dummy_sr)
        csession({'QUERY_STRING':query}, self.dummy_sr)
        csession({'QUERY_STRING':query}, self.dummy_sr)
        result = csession({'QUERY_STRING':query}, self.dummy_sr)
        self.assertEqual(result['count'], 4)

    def test_dec_urlsession_mdc(self):
        '''Tests URL encoded sessions with MemCached decorator.'''
        @memcached.urlsession('localhost')
        def csession(environ, start_response):
            session = environ['com.saddi.service.session'].session
            count = session.get('count', 0) + 1
            session['count'] = count
            environ['count'] = count
            headers = start_response('200 OK', [])
            if headers: environ['cookie'] = headers[0][1]
            return environ
        url = csession({}, self.dummy_sr)[0].split()[-1]
        query = urlparse.urlsplit(url)[3]
        csession({'QUERY_STRING':query}, self.dummy_sr)
        csession({'QUERY_STRING':query}, self.dummy_sr)
        csession({'QUERY_STRING':query}, self.dummy_sr)
        result = csession({'QUERY_STRING':query}, self.dummy_sr)
        self.assertEqual(result['count'], 4)

    def test_wsgimemoize_default_sc(self):
        '''Tests default memoizing with SimpleCache decorator.'''
        testc = simple.SimpleCache()
        env = {'PATH_INFO':'/', 'REQUEST_METHOD':'GET'}
        cacheapp = cache.WsgiMemoize(self.my_app3, testc)
        result1 = cacheapp(env, self.dummy_sr)
        result2 = cacheapp(env, self.dummy_sr)
        self.assertEqual(result1 == result2, True)

    def test_wsgimemoize_default_mc(self):
        '''Tests default memoizing with MemoryCache.'''
        testc = memory.MemoryCache()
        env = {'PATH_INFO':'/', 'REQUEST_METHOD':'GET'}
        cacheapp = cache.WsgiMemoize(self.my_app3, testc)
        result1 = cacheapp(env, self.dummy_sr)
        result2 = cacheapp(env, self.dummy_sr)
        self.assertEqual(result1 == result2, True)

    def test_wsgimemoize_default_fc(self):
        '''Tests default memoizing with FileCache.'''
        testc = file.FileCache('test_wsgistate')
        env = {'PATH_INFO':'/', 'REQUEST_METHOD':'GET'}
        cacheapp = cache.WsgiMemoize(self.my_app3, testc)
        result1 = cacheapp(env, self.dummy_sr)
        result2 = cacheapp(env, self.dummy_sr)
        self.assertEqual(result1 == result2, True)

    def test_wsgimemoize_default_db(self):
        '''Tests default memoizing with DbCache.'''
        testc = db.DbCache('sqlite://')
        env = {'PATH_INFO':'/', 'REQUEST_METHOD':'GET'}
        cacheapp = cache.WsgiMemoize(self.my_app3, testc)
        result1 = cacheapp(env, self.dummy_sr)
        result2 = cacheapp(env, self.dummy_sr)
        self.assertEqual(result1 == result2, True)

    def test_wsgimemoize_default_mcd(self):
        '''Tests default memoizing with MemCached.'''
        testc = memcached.MemCached('localhost')
        env = {'PATH_INFO':'/', 'REQUEST_METHOD':'GET'}
        cacheapp = cache.WsgiMemoize(self.my_app3, testc)
        result1 = cacheapp(env, self.dummy_sr)
        result2 = cacheapp(env, self.dummy_sr)
        self.assertEqual(result1 == result2, True)

    def test_dec_wsgimemoize_default_sc(self):
        '''Tests default memoizing with SimpleCache decorator.'''
        @simple.memoize()
        def cacheapp(environ, start_response):
            start_response('200 OK', [])
            return ['passed']
        env = {'PATH_INFO':'/', 'REQUEST_METHOD':'GET'}
        result1 = cacheapp(env, self.dummy_sr)
        result2 = cacheapp(env, self.dummy_sr)
        self.assertEqual(result1 == result2, True)

    def test_dec_wsgimemoize_default_mc(self):
        '''Tests default memoizing with MemoryCache.'''
        @memory.memoize()
        def cacheapp(environ, start_response):
            start_response('200 OK', [])
            return ['passed']
        env = {'PATH_INFO':'/', 'REQUEST_METHOD':'GET'}
        result1 = cacheapp(env, self.dummy_sr)
        result2 = cacheapp(env, self.dummy_sr)
        self.assertEqual(result1 == result2, True)

    def test_dec_wsgimemoize_default_fc(self):
        '''Tests default memoizing with FileCache.'''
        @file.memoize('test_wsgistate')
        def cacheapp(environ, start_response):
            start_response('200 OK', [])
            return ['passed']
        env = {'PATH_INFO':'/', 'REQUEST_METHOD':'GET'}
        result1 = cacheapp(env, self.dummy_sr)
        result2 = cacheapp(env, self.dummy_sr)
        self.assertEqual(result1 == result2, True)

    def test_dec_wsgimemoize_default_db(self):
        '''Tests default memoizing with DbCache.'''
        @db.memoize('sqlite://')
        def cacheapp(environ, start_response):
            start_response('200 OK', [])
            return ['passed']
        env = {'PATH_INFO':'/', 'REQUEST_METHOD':'GET'}
        result1 = cacheapp(env, self.dummy_sr)
        result2 = cacheapp(env, self.dummy_sr)
        self.assertEqual(result1 == result2, True)

    def test_dec_wsgimemoize_fault_mcd(self):
        '''Tests default memoizing with MemCached.'''
        @memcached.memoize('localhost')
        def cacheapp(environ, start_response):
            start_response('200 OK', [])
            return ['passed']
        env = {'PATH_INFO':'/', 'REQUEST_METHOD':'GET'}
        result1 = cacheapp(env, self.dummy_sr)
        result2 = cacheapp(env, self.dummy_sr)
        self.assertEqual(result1 == result2, True)

    def test_wsgimemoize_method_sc(self):
        '''Tests memoizing with HTTP method as part of the key with SimpleCache.'''
        testc = simple.SimpleCache()
        env = {'PATH_INFO':'/', 'REQUEST_METHOD':'GET'}
        cacheapp = cache.WsgiMemoize(self.my_app3, testc, key_methods=True)
        result1 = cacheapp(env, self.dummy_sr)
        result2 = cacheapp(env, self.dummy_sr)
        self.assertEqual(result1 == result2, True)

    def test_wsgimemoize_method_mc(self):
        '''Tests memoizing with HTTP method as part of the key with MemoryCache.'''
        testc = memory.MemoryCache()
        env = {'PATH_INFO':'/', 'REQUEST_METHOD':'GET'}
        cacheapp = cache.WsgiMemoize(self.my_app3, testc, key_methods=True)
        result1 = cacheapp(env, self.dummy_sr)
        result2 = cacheapp(env, self.dummy_sr)
        self.assertEqual(result1 == result2, True)

    def test_wsgimemoize_method_fc(self):
        '''Tests memoizing with HTTP method as part of the key with FileCache.'''
        testc = file.FileCache('test_wsgistate')
        env = {'PATH_INFO':'/', 'REQUEST_METHOD':'GET'}
        cacheapp = cache.WsgiMemoize(self.my_app3, testc, key_methods=True)
        result1 = cacheapp(env, self.dummy_sr)
        result2 = cacheapp(env, self.dummy_sr)
        self.assertEqual(result1 == result2, True)

    def test_wsgimemoize_method_db(self):
        '''Tests memoizing with HTTP method as part of the key with DbCache.'''
        testc = db.DbCache('sqlite://')
        env = {'PATH_INFO':'/', 'REQUEST_METHOD':'GET'}
        cacheapp = cache.WsgiMemoize(self.my_app3, testc, key_methods=True)
        result1 = cacheapp(env, self.dummy_sr)
        result2 = cacheapp(env, self.dummy_sr)
        self.assertEqual(result1 == result2, True)

    def test_wsgimemoize_method_mcd(self):
        '''Tests memoizing with HTTP method as part of the key with MemCached.'''
        testc = memcached.MemCached('localhost')
        env = {'PATH_INFO':'/', 'REQUEST_METHOD':'GET'}
        cacheapp = cache.WsgiMemoize(self.my_app3, testc, key_methods=True)
        result1 = cacheapp(env, self.dummy_sr)
        result2 = cacheapp(env, self.dummy_sr)
        self.assertEqual(result1 == result2, True)

    def test_wsgimemoize_user_info_sc(self):
        '''Tests memoizing with user info as part of the key with SimpleCache.'''
        testc = simple.SimpleCache()
        env = {'PATH_INFO':'/', 'REQUEST_METHOD':'GET',
        'wsgi.input':StringIO.StringIO('num=12121&str1=test&state=NV&Submit=Submit')}
        cacheapp = cache.WsgiMemoize(self.my_app3, testc, key_user_info=True)
        result1 = cacheapp(env, self.dummy_sr)
        result2 = cacheapp(env, self.dummy_sr)
        self.assertEqual(result1 == result2, True)

    def test_wsgimemoize_user_info_mc(self):
        '''Tests memoizing with user info as part of the key with MemoryCache.'''
        testc = memory.MemoryCache()
        env = {'PATH_INFO':'/', 'REQUEST_METHOD':'GET',
        'wsgi.input':StringIO.StringIO('num=12121&str1=test&state=NV&Submit=Submit')}
        cacheapp = cache.WsgiMemoize(self.my_app3, testc, key_user_info=True)
        result1 = cacheapp(env, self.dummy_sr)
        result2 = cacheapp(env, self.dummy_sr)
        self.assertEqual(result1 == result2, True)

    def test_wsgimemoize_user_info_fc(self):
        '''Tests memoizing with user info as part of the key with FileCache.'''
        testc = file.FileCache('test_wsgistate')
        env = {'PATH_INFO':'/', 'REQUEST_METHOD':'GET',
        'wsgi.input':StringIO.StringIO('num=12121&str1=test&state=NV&Submit=Submit')}
        cacheapp = cache.WsgiMemoize(self.my_app3, testc, key_user_info=True)
        result1 = cacheapp(env, self.dummy_sr)
        result2 = cacheapp(env, self.dummy_sr)
        self.assertEqual(result1 == result2, True)

    def test_wsgimemoize_user_info_db(self):
        '''Tests memoizing with user info as part of the key with DbCache.'''
        testc = db.DbCache('sqlite://')
        env = {'PATH_INFO':'/', 'REQUEST_METHOD':'GET',
        'wsgi.input':StringIO.StringIO('num=12121&str1=test&state=NV&Submit=Submit')}
        cacheapp = cache.WsgiMemoize(self.my_app3, testc, key_user_info=True)
        result1 = cacheapp(env, self.dummy_sr)
        result2 = cacheapp(env, self.dummy_sr)
        self.assertEqual(result1 == result2, True)

    def test_wsgimemoize_user_info_mcd(self):
        '''Tests memoizing with user info as part of the key with MemCached.'''
        testc = memcached.MemCached('localhost')
        env = {'PATH_INFO':'/', 'REQUEST_METHOD':'GET',
        'wsgi.input':StringIO.StringIO('num=12121&str1=test&state=NV&Submit=Submit')}
        cacheapp = cache.WsgiMemoize(self.my_app3, testc, key_user_info=True)
        result1 = cacheapp(env, self.dummy_sr)
        result2 = cacheapp(env, self.dummy_sr)
        self.assertEqual(result1 == result2, True)

    def test_wsgimemoize_allowed_methods(self):
        '''Tests memoizing with HTTP method as part of the key with MemCached.'''
        testc = simple.SimpleCache()
        env = {'PATH_INFO':'/', 'REQUEST_METHOD':'POST',
        'wsgi.input':StringIO.StringIO('num=12121&str1=test&state=NV&Submit=Submit')}
        cacheapp = cache.WsgiMemoize(self.my_app3, testc, allowed_methods=set(['POST']))
        result1 = cacheapp(env, self.dummy_sr)
        result2 = cacheapp(env, self.dummy_sr)
        self.assertEqual(result1 == result2, True)

    def test_wsgimemoize_not_allowed_methods(self):
        '''Tests memoizing with an unallowed HTTP method.'''
        testc = simple.SimpleCache()
        env = {'PATH_INFO':'/', 'REQUEST_METHOD':'POST',
        'wsgi.input':StringIO.StringIO('num=12121&str1=test&state=NV&Submit=Submit')}
        cacheapp = cache.WsgiMemoize(self.my_app3, testc)
        result1 = cacheapp(env, self.dummy_sr)
        result2 = cacheapp(env, self.dummy_sr)
        self.assertEqual(result1 == result2, True)

    def test_public(self):
        '''Tests correct setting of Cache-Control header "public".'''
        @cache.public
        def app(environ, start_response):
            headers = start_response('200 OK', [])
            if headers: environ['headers'] = headers
            return environ
        result = dict(app({'REQUEST_METHOD':'GET'}, self.dummy_sr)['headers'])
        self.assertEqual(result['Cache-Control'], 'public')

    def test_private(self):
        '''Tests correct setting of Cache-Control header "private".'''
        @cache.private
        def app(environ, start_response):
            headers = start_response('200 OK', [])
            if headers: environ['headers'] = headers
            return environ
        result = dict(app({'REQUEST_METHOD':'GET'}, self.dummy_sr)['headers'])
        self.assertEqual(result['Cache-Control'], 'private')

    def test_nocache(self):
        '''Tests correct setting of Cache-Control header "no-cache".'''
        @cache.nocache
        def app(environ, start_response):
            headers = start_response('200 OK', [])
            if headers: environ['headers'] = headers
            return environ
        result = dict(app({'REQUEST_METHOD':'GET'}, self.dummy_sr)['headers'])
        self.assertEqual(result['Cache-Control'], 'no-cache')

    def test_nostore(self):
        '''Tests correct setting of Cache-Control header "no-store".'''
        @cache.nostore
        def app(environ, start_response):
            headers = start_response('200 OK', [])
            if headers: environ['headers'] = headers
            return environ
        result = dict(app({'REQUEST_METHOD':'GET'}, self.dummy_sr)['headers'])
        self.assertEqual(result['Cache-Control'], 'no-store')

    def test_notransform(self):
        '''Tests correct setting of Cache-Control header "no-transform".'''
        @cache.notransform
        def app(environ, start_response):
            headers = start_response('200 OK', [])
            if headers: environ['headers'] = headers
            return environ
        result = dict(app({'REQUEST_METHOD':'GET'}, self.dummy_sr)['headers'])
        self.assertEqual(result['Cache-Control'], 'no-transform')

    def test_revalidate(self):
        '''Tests correct setting of Cache-Control header "must-revalidate".'''
        @cache.revalidate
        def app(environ, start_response):
            headers = start_response('200 OK', [])
            if headers: environ['headers'] = headers
            return environ
        result = dict(app({'REQUEST_METHOD':'GET'}, self.dummy_sr)['headers'])
        self.assertEqual(result['Cache-Control'], 'must-revalidate')

    def test_proxyrevalidate(self):
        '''Tests correct setting of Cache-Control header "proxy-revalidate".'''
        @cache.proxyrevalidate
        def app(environ, start_response):
            headers = start_response('200 OK', [])
            if headers: environ['headers'] = headers
            return environ
        result = dict(app({'REQUEST_METHOD':'GET'}, self.dummy_sr)['headers'])
        self.assertEqual(result['Cache-Control'], 'proxy-revalidate')

    def test_maxage(self):
        '''Tests correct setting of Cache-Control header "max-age".'''
        @cache.maxage(30)
        def app(environ, start_response):
            headers = start_response('200 OK', [])
            if headers: environ['headers'] = headers
            return environ
        result = dict(app({'REQUEST_METHOD':'GET'}, self.dummy_sr)['headers'])
        self.assertEqual(result['Cache-Control'], 'max-age=30')

    def test_smaxage(self):
        '''Tests correct setting of Cache-Control header "s-maxage".'''
        @cache.smaxage(30)
        def app(environ, start_response):
            headers = start_response('200 OK', [])
            if headers: environ['headers'] = headers
            return environ
        result = dict(app({'REQUEST_METHOD':'GET'}, self.dummy_sr)['headers'])
        self.assertEqual(result['Cache-Control'], 's-maxage=30')

    def test_vary(self):
        '''Tests correct setting of Vary header.'''
        @cache.vary(['Content-type'])
        def app(environ, start_response):
            headers = start_response('200 OK', [])
            if headers: environ['headers'] = headers
            return environ
        result = dict(app({'REQUEST_METHOD':'GET'}, self.dummy_sr)['headers'])
        self.assertEqual(result['Vary'], 'Content-type')

    def test_modified(self):
        '''Tests setting of Modified header.'''
        @cache.modified(30)
        def app(environ, start_response):
            headers = start_response('200 OK', [])
            if headers: environ['headers'] = headers
            return environ
        result = dict(app({'REQUEST_METHOD':'GET'}, self.dummy_sr)['headers'])
        self.assertEqual('Modified' in result, True)

    def test_cachecontrol_combo(self):
        '''Tests correct setting of two Cache Control header settings.'''
        @cache.smaxage(30)
        @cache.private
        def app(environ, start_response):
            headers = start_response('200 OK', [])
            if headers: environ['headers'] = headers
            return environ
        result = dict(app({'REQUEST_METHOD':'GET'}, self.dummy_sr)['headers'])
        self.assertEqual(result['Cache-Control'], 's-maxage=30, private')

    'modified'

if __name__ == '__main__':
    unittest.main()
    try:
        os.rmdir('test_wsgistate')
    except IOError: pass


# Copyright (c) 2006-2007 L. C. Rees.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1.  Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# 2.  Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
# 3.  Neither the name of the Portable Site Information Project nor the names
# of its contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

'''setup - setuptools based setup for wsgistate.'''

import ez_setup
ez_setup.use_setuptools()

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

setup(
    name='wsgistate',
    version='0.4.2',
    description='''WSGI session and caching middleware.''',
    long_description='''Session (flup-compatible), caching, memoizing, and HTTP cache control
middleware for WSGI. Supports memory, filesystem, database, and memcached based backends.

# Simple memoization example:

from wsgistate.memory import memoize

@memoize()
def app(environ, start_response):
    start_response('200 OK', [('Content-Type', 'text/plain')])
    return ['Hello World!']

if __name__ == '__main__':
    from wsgiref.simple_server import make_server
    http = make_server('', 8080, app)
    http.serve_forever()

# Simple session example:

from wsgistate.memory import session

@session()
def app(environ, start_response):
    session = environ['com.saddi.service.session'].session
    count = session.get('count', 0) + 1
    session['count'] = count
    start_response('200 OK', [('Content-Type', 'text/plain')])
    return ['You have been here %d times!\n' % count]

if __name__ == '__main__':
    from wsgiref.simple_server import make_server
    http = make_server('', 8080, app)
    http.serve_forever()''',
    author='L. C. Rees',
    author_email='lcrees@gmail.com',
    url='http://pypi.python.org/pypi/wsgistate/',
    license='BSD',
    packages = ['wsgistate', 'wsgistate.tests'],
    test_suite='wsgistate.tests',
    zip_safe = False,
    keywords='WSGI session caching persistence memoizing HTTP Web',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Middleware',
    ],
    install_requires = ['SQLAlchemy>0.3', 'python-memcached'],
    entry_points='''
    [paste.filter_factory]
    file_memo=wsgistate.file:filememo_deploy
    firebird_memo=wsgistate.db:dbmemo_deploy
    memcache_memo=wsgistate.memcached:mcmemo_deploy
    memory_memo=wsgistate.memory:memorymemo_deploy
    mssql_memo=wsgistate.db:dbmemo_deploy
    mysql_memo=wsgistate.db:dbmemo_deploy
    oracle_memo=wsgistate.db:dbmemo_deploy
    postgres_memo=wsgistate.db:dbmemo_deploy
    simple_memo=wsgistate.simple:simplememo_deploy
    sqlite_memo=wsgistate.db:dbmemo_deploy
    file_session=wsgistate.file:filesess_deploy
    firebird_session=wsgistate.db:dbsess_deploy
    memcache_session=wsgistate.memcached:mcsess_deploy
    memory_session=wsgistate.memory:memorysess_deploy
    mssql_session=wsgistate.db:dbsess_deploy
    mysql_session=wsgistate.db:dbsess_deploy
    oracle_session=wsgistate.db:dbsess_deploy
    postgres_session=wsgistate.db:dbsess_deploy
    simple_session=wsgistate.simple:simplesess_deploy
    sqlite_session=wsgistate.db:dbsess_deploy
    file_urlsess=wsgistate.file:fileurlsess_deploy
    firebird_urlsess=wsgistate.db:dburlsess_deploy
    memcache_urlsess=wsgistate.memcached:mcurlsess_deploy
    memory_urlsess=wsgistate.memory:memoryurlsess_deploy
    mssql_urlsess=wsgistate.db:dburlsess_deploy
    mysql_urlsess=wsgistate.db:dburlsess_deploy
    oracle_urlsess=wsgistate.db:dburlsess_deploy
    postgres_urlsess=wsgistate.db:dburlsess_deploy
    simple_urlsess=wsgistate.simple:simpleurlsess_deploy
    sqlite_urlsess=wsgistate.db:dburlsess_deploy
    '''
)


# Copyright (c) 2005, the Lawrence Journal-World
# Copyright (c) 2006 L. C. Rees
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#    1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#    2. Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#
#    3. Neither the name of Django nor the names of its contributors may be used
#       to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Single-process in-memory cache backend.'''

import time
import random
from wsgistate import BaseCache
from wsgistate.cache import WsgiMemoize
from wsgistate.session import CookieSession, URLSession, SessionCache

__all__ = ['SimpleCache', 'memoize', 'session', 'urlsession']

def simplememo_deploy(global_conf, **kw):
    '''Paste Deploy loader for caching.'''
    def decorator(application):
        _simple_memo_cache = SimpleCache(kw.get('cache'), **kw)
        return WsgiMemoize(application, _simple_memo_cache, **kw)
    return decorator

def simplesess_deploy(global_conf, **kw):
    '''Paste Deploy loader for sessions.'''
    def decorator(application):
        _simple_base_cache = SimpleCache(kw.get('cache'), **kw)
        _simple_session_cache = SessionCache(_simple_base_cache, **kw)
        return CookieSession(application, _simple_session_cache, **kw)
    return decorator

def simpleurlsess_deploy(global_conf, **kw):
    '''Paste Deploy loader for URL encoded sessions.

    @param initstr Database initialization string
    '''
    def decorator(application):
        _simple_ubase_cache = SimpleCache(kw.get('cache'), **kw)
        _simple_url_cache = SessionCache(_simple_ubase_cache, **kw)
        return URLSession(application, _simple_url_cache, **kw)
    return decorator

def memoize(**kw):
    '''Decorator for caching.'''
    def decorator(application):
        _simple_memo_cache = SimpleCache(**kw)
        return WsgiMemoize(application, _simple_memo_cache, **kw)
    return decorator

def session(**kw):
    '''Decorator for sessions.'''
    def decorator(application):
        _simple_base_cache = SimpleCache(**kw)
        _simple_session_cache = SessionCache(_simple_base_cache, **kw)
        return CookieSession(application, _simple_session_cache, **kw)
    return decorator

def urlsession(**kw):
    '''Decorator for URL encoded sessions.'''
    def decorator(application):
        _simple_ubase_cache = SimpleCache(**kw)
        _simple_url_cache = SessionCache(_simple_ubase_cache, **kw)
        return URLSession(application, _simple_url_cache, **kw)
    return decorator


class SimpleCache(BaseCache):

    '''Single-process in-memory cache backend.'''

    def __init__(self, *a, **kw):
        super(SimpleCache, self).__init__(*a, **kw)
        # Get random seed
        random.seed()
        self._cache = dict()
        # Set max entries
        max_entries = kw.get('max_entries', 300)
        try:
            self._max_entries = int(max_entries)
        except (ValueError, TypeError):
            self._max_entries = 300
        # Set maximum number of items to cull if over max
        self._maxcull = kw.get('maxcull', 10)

    def get(self, key, default=None):
        '''Fetch a given key from the cache.  If the key does not exist, return
        default, which itself defaults to None.

        @param key Keyword of item in cache.
        @param default Default value (default: None)
        '''
        values = self._cache.get(key)
        if values is None: return default
        # Delete if item timed out and return default.
        if values[0] < time.time():
            self.delete(key)
            return default
        return values[1]

    def set(self, key, value):
        '''Set a value in the cache.

        @param key Keyword of item in cache.
        @param value Value to be inserted in cache.
        '''
        # Cull timed out values if over max # of entries
        if len(self._cache) >= self._max_entries: self._cull()
        # Set value and timeout in cache
        self._cache[key] = (time.time() + self.timeout, value)

    def delete(self, key):
        '''Delete a key from the cache, failing silently.

        @param key Keyword of item in cache.
        '''
        try:
            del self._cache[key]
        except KeyError: pass

    def keys(self):
        '''Returns a list of keys in the cache.'''
        return self._cache.keys()

    def _cull(self):
        '''Remove items in cache to make room.'''
        num, maxcull = 0, self._maxcull
        # Cull number of items allowed (set by self._maxcull)
        for key in self.keys():
            # Remove only maximum # of items allowed by maxcull
            if num <= maxcull:
                # Remove items if expired
                if self.get(key) is None: num += 1
            else: break
        # Remove any additional items up to max # of items allowed by maxcull
        while len(self.keys()) >= self._max_entries and num <= maxcull:
            # Cull remainder of allowed quota at random
            self.delete(random.choice(self.keys()))
            num += 1